; Example Program seeprom01.a51
; Serial EEPROM (AT24Cxx)
; Chin-Shiuh Shieh
; 2003/03/22
;
PROG			EQU	0000h
;
; Pin Assignment
SEEPROM_SCL		EQU	P1.0	; Serial Clock Input
SEEPROM_SDA		EQU	P1.1	; Serial Data
;
; Memory Allocation
SEEPROM_MADRS_LO	EQU	30h	; High Address
SEEPROM_MADRS_HI	EQU	31h	; Low Address
SEEPROM_DATA		EQU	32h	; Data
SEEPROM_SBUF		EQU	33h	; Shift Buffer
UART_BUSY		EQU	00h	; UART Busy Flag Bit
;
; Symbol Definition
SEEPROM_DADRS		EQU	0A0h
;
	ORG		PROG+0000h
	SJMP	START
;Interrupt Vector Table
	ORG	PROG+0003h
	LCALL	INT0_ISR
	RETI
	ORG	PROG+000Bh
	LCALL	T0_ISR
	RETI
	ORG	PROG+0013h
	LCALL	INT1_ISR
	RETI
	ORG	PROG+001Bh
	LCALL	T1_ISR
	RETI
	ORG	PROG+0023h
	LCALL	UART_ISR
	RETI
;
	ORG	PROG+0030h

START:
        MOV     TMOD,#00100001b	; Timer1 in Mode 2, Timer0 in Mode 1
        MOV     TH1, #0FDh	; Baud Rate = 9600 bps at 11.0592MHz
        MOV     SCON,#01010000b	; UART in Mode 1
	SETB	ES		; Enable UART Interrupt
	SETB	EA		; Enable Interrupt
	SETB	TR1		; Start Timer 1
	CLR	UART_BUSY
; Write "ABC..." into SEEPROM
	MOV	SEEPROM_MADRS_HI,#00h
	MOV	SEEPROM_MADRS_LO,#00h
	MOV	SEEPROM_DATA,#'A'
LOOP_1:	LCALL	SEEPROM_WRITE
	INC	SEEPROM_DATA
	MOV	A,SEEPROM_DATA
	CJNE	A,#'Z'+1,INC_MADRS_LO_1
	MOV	SEEPROM_DATA,#'A'
INC_MADRS_LO_1:
	INC	SEEPROM_MADRS_LO
	MOV	A,SEEPROM_MADRS_LO
	CJNE	A,#00h,LOOP_1
INC_MADRS_HI_1:
	INC	SEEPROM_MADRS_HI
	MOV	A,SEEPROM_MADRS_HI
	CJNE	A,#20h,LOOP_1		; 8K Limit for AT24C64
; Read bytes from SEEPROM and sent to UART
	MOV	SEEPROM_MADRS_HI,#00h
	MOV	SEEPROM_MADRS_LO,#00h
LOOP_2:	LCALL	SEEPROM_READ
WAIT_UART:
	JB	UART_BUSY,WAIT_UART
	SETB	UART_BUSY
	MOV	SBUF,SEEPROM_DATA
INC_MADRS_LO_2:
	INC	SEEPROM_MADRS_LO
	MOV	A,SEEPROM_MADRS_LO
	CJNE	A,#00h,LOOP_2
INC_MADRS_HI_2:
	INC	SEEPROM_MADRS_HI
	MOV	A,SEEPROM_MADRS_HI
	CJNE	A,#20h,LOOP_2		; 8K Limit for AT24C64
DONE:	SJMP	DONE
;
INT0_ISR:
	RET
T0_ISR:
	RET
INT1_ISR:
	RET
T1_ISR:
	RET
UART_ISR:
        JB      RI,RECEIVED
TRANSMITTED:
	CLR	UART_BUSY
        CLR     TI
        RET
RECEIVED:
	CLR     RI
	RET
;
SEEPROM_WRITE:
; Serial EEPROM (AT24Cxx) Byte Write function.
; Device address: SEEPROM_DADRS
; Byte address :  SEEPROM_MADRS_HI and SEEPROM_MADRS_LO
; Data: 	  SEEPROM_DATA
; CY = 1 if bus is not available or addressed device failed to acknowledge.
; Retry if CY = 1.
	PUSH	ACC
	LCALL	SEEPROM_START
	JC	SEEPROM_WRITE_RET		; abort if bus not available
	MOV	A,#SEEPROM_DADRS		; specify device address
	CLR	ACC.0				; specify write operation
	MOV	SEEPROM_SBUF,A
	LCALL	SEEPROM_SHOUT			; send device address
	JC	SEEPROM_WRITE_ERROR		; abort if no acknowledge
	MOV	SEEPROM_SBUF,SEEPROM_MADRS_HI	; specify high byte of memory address
	LCALL	SEEPROM_SHOUT			; send high byte of memory address
	JC	SEEPROM_WRITE_ERROR		; abort if no acknowledge
	MOV	SEEPROM_SBUF,SEEPROM_MADRS_LO	; specify low byte of memory address
	LCALL	SEEPROM_SHOUT			; send low byte of memory address
	JC	SEEPROM_WRITE_ERROR		; abort if no acknowledge
	MOV	SEEPROM_SBUF,SEEPROM_DATA	; specify data
	LCALL	SEEPROM_SHOUT			; send data
	JC	SEEPROM_WRITE_ERROR		; abort if no acknowledge
	CLR	C				; clear error flag
SEEPROM_WRITE_ERROR:
	LCALL	SEEPROM_STOP
SEEPROM_WRITE_RET:
	POP	ACC
	JC	SEEPROM_WRITE			; retry if error
	RET
;
SEEPROM_READ:
; Serial EEPROM (AT24Cxx) Random Read function.
; Device address: SEEPROM_DADRS
; Byte address :  SEEPROM_MADRS_HI and SEEPROM_MADRS_LO
; Data: 	  SEEPROM_DATA
; CY = 1 if bus is not available addressed device failed to acknowledge.
; Retry if CY = 1.
	PUSH	ACC
	LCALL	SEEPROM_START
	JC	SEEPROM_READ_RET		; abort if bus not available
	MOV	A,#SEEPROM_DADRS		; specify device address
	CLR	ACC.0				; specify write operation
	MOV	SEEPROM_SBUF,A
	LCALL	SEEPROM_SHOUT			; send device address
	JC	SEEPROM_READ_ERROR		; abort if no acknowledge
	MOV	SEEPROM_SBUF,SEEPROM_MADRS_HI	; specify high byte of memory address
	LCALL	SEEPROM_SHOUT			; send high byte of memory address
	JC	SEEPROM_READ_ERROR		; abort if no acknowledge
	MOV	SEEPROM_SBUF,SEEPROM_MADRS_LO	; specify low byte of memory address
	LCALL	SEEPROM_SHOUT			; send low byte of memory address
	JC	SEEPROM_READ_ERROR		; abort if no acknowledge
	LCALL	SEEPROM_START
	JC	SEEPROM_READ_RET		; abort if bus not available
	MOV	A,#SEEPROM_DADRS		; specify device address
	SETB	ACC.0				; specify read operation
	MOV	SEEPROM_SBUF,A
	LCALL	SEEPROM_SHOUT			; send device address
	JC	SEEPROM_READ_ERROR		; abort if no acknowledge
	LCALL	SEEPROM_SHIN			; receive data byte
	LCALL	SEEPROM_NAK			; do not acknowledge byte
	MOV	SEEPROM_DATA,SEEPROM_SBUF	; store received data
	CLR	C				; clear error flag
SEEPROM_READ_ERROR:
	LCALL	SEEPROM_STOP
SEEPROM_READ_RET:
	POP	ACC
	JC	SEEPROM_READ			; retry if error
	RET
;
SEEPROM_START:
; Send START, defined as high-to-low SDA with SCL high.
; Return with SCL, SDA low.
; Returns CY set if bus is not available.
	SETB	SEEPROM_SDA
	SETB	SEEPROM_SCL
	JNB	SEEPROM_SDA,SEEPROM_START_ERROR
	JNB	SEEPROM_SCL,SEEPROM_START_ERROR
	NOP			; enforce setup delay and cycle delay
	CLR	SEEPROM_SDA
	NOP			; enforce hold delay
	NOP
	NOP
	NOP
	NOP
	CLR	SEEPROM_SCL
	CLR	C		; clear error flag
	SJMP	SEEPROM_START_RET
SEEPROM_START_ERROR:
		SETB	C	; set error flag
SEEPROM_START_RET:
		RET
;
SEEPROM_STOP:
; Send STOP, defined as low-to-high SDA with SCL high.
; SCL expected low on entry. Return with SCL, SDA high.
	CLR	SEEPROM_SDA
	NOP			; enforce SCL low and data setup
	NOP
	SETB	SEEPROM_SCL
	NOP			; enforce setup delay
	NOP
	NOP
	NOP
	NOP
	SETB	SEEPROM_SDA
	RET
;
SEEPROM_SHOUT:
; Shift out a byte to the Serial EPROM (AT24C64), most significant bit first.
; SCL, SDA expected low on entry. Return with SCL low.
; Called with data to send in SEEPROM_SBUF.
; Returns CY set to indicate failure by slave to acknowledge.
	PUSH	ACC
	PUSH	B
	MOV	A,SEEPROM_SBUF
	MOV	B,#8			; bit counter
SEEPROM_SHOUT_LOOP:
	RLC	A			; move bit into CY
	MOV	SEEPROM_SDA,C		; output bit
	NOP				; enforce SCL low and data setup
	SETB	SEEPROM_SCL		; raise clock
	NOP				; enforce SCL high
	NOP
	NOP
	NOP
	CLR	SEEPROM_SCL		; drop clock
	DJNZ	B,SEEPROM_SHOUT_LOOP
	SETB	SEEPROM_SDA		; release SDA for ACK
	NOP				; enforce SCL low and tAA
	NOP
	SETB	SEEPROM_SCL		; raise ACK clock
	NOP				; enforce SCL high
	NOP
	NOP
	NOP
	MOV	C,SEEPROM_SDA		; get ACK bit
	CLR	SEEPROM_SCL		; drop ACK clock
	POP	B
	POP	ACC
	RET
;
SEEPROM_SHIN:
; Shift in a byte from the Serial EEPROM(AT24C64), most significant bit first.
; SCL expected low on entry. Return with SCL low.
; Returns received data byte in SEEPROM_SBUF.
	PUSH	ACC
	PUSH	B
	SETB	SEEPROM_SDA		; make SDA an input
	MOV	B,#8			; bit count
SEEPROM_SHIN_LOOP:
	NOP				; enforce SCL low and data setup
	NOP
	NOP
	SETB	SEEPROM_SCL		; raise clock
	NOP				; enforce SCL high
	NOP
	MOV	C, SEEPROM_SDA		; input bit
	RLC	A			; move bit into byte
	CLR	SEEPROM_SCL		; drop clock
	DJNZ	B,SEEPROM_SHIN_LOOP
	MOV	SEEPROM_SBUF,A
	POP	B
	POP	ACC
	RET
;
SEEPROM_NAK:
; Clock out a negative acknowledge bit (high).
; SCL expected low on entry. Return with SCL low, SDA high.
	SETB	SEEPROM_SDA	; NAK bit
	NOP			; enforce SCL low and data setup
	NOP
	SETB	SEEPROM_SCL	; raise clock
	NOP			; enforce SCL high
	NOP
	NOP
	NOP
	CLR	SEEPROM_SCL	; drop clock
	RET
;
		END
